#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <glib.h>

#include <g3dviewer_plugin.h>

#define MD2_SKINNAMELEN 64

int plugin_load(const char *filename, G3DModel *model)
{
  FILE *f;
  guint32 idid, idver, skinwidth, skinheight, framesize;
  guint32 numskins, numverts, numtexs, numfaces, numglcmds, numframes;
  guint32 offskins, offtexs, offfaces, offframes, offglcmds, offend;
  char **skinnames;
  int i;
  G3DObject *object;
  G3DMaterial *material;

  f = fopen(filename, "r");
  if(f == NULL)
  {
    g_printerr("couldn't open '%s'\n", filename);
    return EXIT_FAILURE;
  }
  
  idid = g3d_read_int32_be(f);
  if(idid != G3D_IFF_MKID('I','D','P','2'))
  {
    g_printerr("file '%s' is not a .md2 file\n", filename);
    fclose(f);
    return EXIT_FAILURE;
  }

  idver = g3d_read_int32_le(f);
  if(idver != 8)
  {
    g_printerr("file '%s' has wrong version (%d)\n", filename, idver);
    fclose(f);
    return EXIT_FAILURE;
  }

  object = g_new0(G3DObject, 1);
  object->name = g_strdup("Q2Object");
  material = g3d_new_G3DMaterial();
  object->materials = g_slist_append(object->materials, material);
  model->objects = g_slist_append(model->objects, object);

  skinwidth  = g3d_read_int32_le(f);
  skinheight = g3d_read_int32_le(f);
  framesize  = g3d_read_int32_le(f);
  numskins   = g3d_read_int32_le(f);
  numverts   = g3d_read_int32_le(f);
  numtexs    = g3d_read_int32_le(f);
  numfaces   = g3d_read_int32_le(f);
  numglcmds  = g3d_read_int32_le(f);
  numframes  = g3d_read_int32_le(f);

  object->vertex_count = numverts;
  object->vertices = g_malloc(numverts * 3 * sizeof(GLfloat));
  
  offskins   = g3d_read_int32_le(f);
  offtexs    = g3d_read_int32_le(f);
  offfaces   = g3d_read_int32_le(f);
  offframes  = g3d_read_int32_le(f);
  offglcmds  = g3d_read_int32_le(f);
  offend     = g3d_read_int32_le(f);

  skinnames = g_malloc0(sizeof(char*) * numskins);
  for(i=0; i<numskins; i++)
  {
    skinnames[i] = g_malloc0(sizeof(char) * MD2_SKINNAMELEN);
    fread(skinnames[i], 1, MD2_SKINNAMELEN, f);
#if DEBUG > 0
    g_printerr("skin #%d: %s\n", i+1, skinnames[i]);
#endif
  }

  /* texture coordinates */
  for(i=0; i<numtexs; i++)
  {
    int s, t;

    s = g3d_read_int16_le(f);
    t = g3d_read_int16_le(f);
  }

  /* faces */
  for(i=0; i<numfaces; i++)
  {
    G3DFace *face;
    
    face = g_new0(G3DFace, 1);
    object->faces = g_slist_append(object->faces, face);
    face->material = material;
    face->num_vertices = 3;
    face->index_vertex = g_malloc(3 * sizeof(int));
    
    face->index_vertex[0] = g3d_read_int16_le(f);
    face->index_vertex[1] = g3d_read_int16_le(f);
    face->index_vertex[2] = g3d_read_int16_le(f);
    g3d_read_int16_le(f);
    g3d_read_int16_le(f);
    g3d_read_int16_le(f);
  }

  /* vertices per frame */
#if DEBUG > 0
  g_printerr("numframes: %d\n", numframes);
#endif
  for(i=0; i<numframes; i++)
  {
    float s0,s1,s2, t0,t1,t2;
    char fname[16];
    int j;

    s0 = g3d_read_float_le(f); /* scale */
    s1 = g3d_read_float_le(f);
    s2 = g3d_read_float_le(f);
    t0 = g3d_read_float_le(f); /* translate */
    t1 = g3d_read_float_le(f);
    t2 = g3d_read_float_le(f);
    fread(fname, 1, 16, f); /* frame name*/

    for(j=0; j<numverts; j++)
    {
      GLfloat x,y,z;
      unsigned char v,n;

      v = g3d_read_int8(f);
      x = (float)v * s0 + t0;
      v = g3d_read_int8(f);
      y = (float)v * s1 + t1;
      v = g3d_read_int8(f);
      z = (float)v * s2 + t2;
      n = g3d_read_int8(f);
      if(i == 0)
      {
        object->vertices[j*3+0] = x;
        object->vertices[j*3+1] = y;
        object->vertices[j*3+2] = z;
      }
    }
  }

  /* free skin names */
  for(i=0; i<numskins; i++) g_free(skinnames[i]);
  g_free(skinnames);

  fclose(f);
  return EXIT_SUCCESS;
}

char *plugin_description(void)
{
  return g_strdup(
    "import plugin for ID Software's Quake II models\n");
}

char **plugin_extensions(void)
{
  return g_strsplit("md2", ":", 0);
}

